/**
 * 基于数据库元信息的序列化和反序列化
 */

import { BonEncode, BonDecode, BonBuffer } from "./bon";
import { EnumType, Type } from "./sinfo";




/**
 * 序列化数据库
 */
export class TabEncode implements BonEncode {
    meta: EnumType;      // 元信息
    data: any;

    constructor(meta: EnumType, data?: any) {
        this.meta = meta;
        this.data = data;
    }
    bonEncode(bb: BonBuffer, sub = false): BonBuffer {
        let type = this.meta.type
        switch (type) {
            case Type.Arr:
                bb.writeArray(this.data, (el) => {
                    new TabEncode(this.meta.into, el).bonEncode(bb, true);
                })
                break;
            case Type.Option:
                if (this.data === undefined || this.data === null) {
                    bb.writeNil();
                } else {
                    new TabEncode(this.meta.into, this.data).bonEncode(bb, true);
                }
                break;
            case Type.Map:
                bb.writeMap(this.data, (k, v) => {
                    new TabEncode(this.meta.mapType[0], k).bonEncode(bb, true);
                    new TabEncode(this.meta.mapType[1], v).bonEncode(bb, true);
                });
                break;
            case Type.Struct:
                if (!sub) {
                    bb.writeCt(this.data, (bb, _o) => {
                        bb.writeU32(this.meta.structType.name_hash); // 写容器类型
                        for (const fieldInfo of this.meta.structType.fields) {
                            new TabEncode(fieldInfo.ftype, this.data[fieldInfo.name]).bonEncode(bb, true);
                        }
                    });
                } else {
                    for (const fieldInfo of this.meta.structType.fields) {
                        new TabEncode(fieldInfo.ftype, this.data[fieldInfo.name]).bonEncode(bb, true);
                    }
                }
                break;
            case Type.Enum:
                bb.writeInt(this.data);
                break;
            case Type.U8: case Type.U16: case Type.U32: case Type.Usize: case Type.I8: case Type.I16: case Type.I32: case Type.Isize:
                bb.writeInt(this.data);
                break;
            case Type.U64: case Type.U128: case Type.U256: case Type.I64: case Type.I128: case Type.I256:
                // TODO
                break;
            case Type.F32:
                bb.writeF32(this.data);
                break;
            case Type.F64:
                bb.writeF64(this.data);
                break;
            case Type.Str:
                bb.writeUtf8(this.data);
                break;
            case Type.Bin:
                bb.writeBin(this.data);
                break;
            case Type.Bool:
                bb.writeBool(this.data);
                break;
            default:
                break;
        }

        return bb;
    }
}


/**
 * 反序列化数据库
 */
export class TabDecode implements BonDecode<any> {
    meta: EnumType;      // 元信息
    data: any;

    constructor(meta: EnumType) {
        this.meta = meta;
    }
    bonDecode(bb: BonBuffer, sub = false): any {
        let type = this.meta.type
        // 	0 表示忽略
        // 	1 通用对象
        // 	2 通用数组
        // 	3 通用map
        let t;
        switch (type) {
            case Type.Arr:
                this.data = bb.readArray(() => {
                    return new TabDecode(this.meta.into).bonDecode(bb, true);
                })
                break;
            case Type.Option:
                if (!bb.isNil()) {
                    this.data = new TabDecode(this.meta.into).bonDecode(bb, true);
                }
                break;
            case Type.Map:
                this.data = bb.readMap(() => {
                    return [new TabDecode(this.meta.mapType[0]).bonDecode(bb, true), new TabDecode(this.meta.mapType[1]).bonDecode(bb, true)];
                });
                break;
            case Type.Struct:
                this.data = {};
                if (!sub) {
                    t = bb.readU8();
                    // 245=8位长度的容器，246=16位长度的容器，247=32位长度的容器，248=48位长度的容器
                    if (t === 245) {
                        bb.head += 1;
                    } else if (t === 246) {
                        bb.head += 2;
                    } else if (t === 247) {
                        bb.head += 4;
                    } else if (t === 248) {
                        bb.head += 6;
                    } else {
                        throw new Error("bon type error type = " + t);
                    }
                    let hash = bb.readU32();    // 结构体hash
                    if (hash !== this.meta.structType.name_hash) throw new Error("name_hash error hash:" + hash);
                }
                for (const fieldInfo of this.meta.structType.fields) {
                    this.data[fieldInfo.name] = new TabDecode(fieldInfo.ftype).bonDecode(bb, true);
                }
                break;
            case Type.Enum:
                this.data = bb.readInt();
                break;
            case Type.U8: case Type.U16: case Type.U32: case Type.Usize: case Type.I8: case Type.I16: case Type.I32: case Type.Isize:
                this.data = bb.readInt();
                break;
            case Type.U64: case Type.U128: case Type.U256: case Type.I64: case Type.I128: case Type.I256:
                // TODO
                break;
            case Type.F32:
                this.data = bb.readf()
                break;
            case Type.F64:
                this.data = bb.readf()
                break;
            case Type.Str:
                this.data = bb.readUtf8();
                break;
            case Type.Bin:
                this.data = bb.readBin();
                break;
            case Type.Bool:
                this.data = bb.readBool();
                break;
            default:
                break;
        }

        return this.data;
    }
}