/**
 * 提供 配置数据 的管理
 * @packageDocumentation
 */

import { baseType } from "../lang/type";
import { Struct } from "../serialization/struct_mgr";

/**
 * Cfg 配置管理对象
 */
export class Cfg {
    /**
     * 数据表
     */
    public map = new Map<string, Map<any, Struct>>();

    /**
     * 获取 primary 表
     * @param {string} key key
     * @param {string} primary primary
     * @returns {Map<baseType, Struct>} 
     */
    getPrimary(key: string, primary: string): Map<baseType, Struct> {
        return this.map.get(`${key}#${primary}`);
    }

    get(key: string): Map<baseType, Struct> {
        return this.map.get(key);
    }

    set(key: string, value: Map<any, Struct>) {
        this.map.set(key, value);
        let notes = (<any>value.get(0).constructor)._$info.notes;
        if (notes) {
            let primary = notes.get("primary");
            if (primary) {
                let primarys = primary.split("-");
                for (let i = 0; i < primarys.length; i++) {
                    let primaryMap = new Map<any, Struct>();
                    value.forEach((v, _k) => {
                        primaryMap.set(v[primarys[i]], v);
                    });
                    this.map.set(`${key}#${primarys[i]}`, primaryMap);
                }
            }
        }
    }

    setArr(key: string, value: Struct[]) {
        // this.map.set(key, value);
        if (value.length > 0) {
            let notes = (<any>value[0].constructor)._$info.notes;
            if (notes) {
                let primary = notes.get("primary");
                if (primary) {
                    let primarys = primary.split("-");
                    for (let i = 0; i < primarys.length; i++) {
                        let primaryMap = new Map<any, Struct>();
                        for (let j = 0; j < value.length; j++) {
                            let v = value[j];
                            primaryMap.set(v[primarys[i]], v);
                        }
                        this.map.set(`${key}#${primarys[i]}`, primaryMap);
                    }
                    return;
                }
            }

            let primaryMap = new Map<number, Struct>();
            for (let j = 0; j < value.length; j++) {
                let v = value[j];
                primaryMap.set(j, v);
            }
            this.map.set(`${key}`, primaryMap);
        }
    }

    update(key: string, value: Map<baseType, Struct>) {
        if (value.size === 0) {
            return;
        }
        let m = this.map.get(key);
        if (!m) {
            this.set(key, value);
        } else {
            let size = m.size;
            value.forEach((v, _) => {
                m.set(size++, v);
            });
            if (value.size === 0) {
                return;
            }
            let notes = (<any>value.get(0).constructor)._$info.notes;
            if (notes) {
                let primary = notes.get("primary");
                if (primary) {
                    let primarys = primary.split("-");
                    for (let i = 0; i < primarys.length; i++) {
                        let map = this.map.get(`${key}#${primarys[i]}`);
                        value.forEach((v, _) => {
                            map.set(v[primarys[i]], v);
                        });
                    }
                }
            }
        }
    }
}

/**
 * 全局Cfg配置的管理器对象
 */
export const cfgMgr = new Cfg();


/**
 * 解析 Jcfg，Jcfg就是cfg内容的合并版本
 * @param cfgs jcfg的json对象的内容，{每个cfg路径: cfg内容}
 */
export const parseJCfg = (cfgs: any): Promise<any> => {
    let rs = [];
    for (let k in cfgs) {
        rs.push(parseCfg(cfgs[k]));
    }
    return Promise.all(rs);
}

/**
 * 解析 cfg内容，放到 cfgMgr 给项目 使用；
 * @param cfgs jcfg的json对象的内容，{每个cfg路径: cfg内容}
 */
export const parseCfg = (cfgs: any): Promise<any> => {
    try {
        let all = [];
        for (let key in cfgs) {
            let file = cfgs[key];
            let split_index = key.indexOf(".");
            let modName = key.slice(0, split_index) + ".struct";
            let Ty = key.slice(split_index + 1, key.length);
            let r = import(modName).then((mod) => {
                let m = mod[Ty];
                let _$c = (path, notes) => { return new m(path, notes) };
                try {
                    let i = 0;
                    let flag = false;
                    let arr = [];
                    for (var v of file) {
                        if (m.prototype.__create) {
                            arr.push(m.prototype.__create(v));
                            flag = true;
                        } else {
                            arr.push([i, _$c(v[0], new Map<string, string>(v[1]))]);
                        }
                        i += 1;
                    }
                    if (flag) {
                        cfgMgr.setArr(key, arr);
                    } else {
                        cfgMgr.update(key, new Map<number, any>(arr));
                    }
                } catch (e) {
                    throw e;
                }
            });
            all.push(r);
        }
        return Promise.all(all);
    } catch (e) {
        return Promise.reject(e);
    }
}

/**
 * @param cfgsstring 
 * Erlang版本的配置解析，因为Erlang只能使用json字符串；
 * 主要用于三国项目；
 */
export const parseErlJCfg = (cfgsstring: any): Promise<any> => {
    var cfgs = JSON.parse(cfgsstring);
    let rs = [];
    for (let k in cfgs) {
        rs.push(parseCfg(cfgs[k]));
    }
    return Promise.all(rs);
}