/**
 * 克隆对象及装饰器方法
 */
/**
 * Clone接口
 */
export interface Clone {
	clone():this
}

/**
 * 类装饰器， 为指定的类生成clone方法， 默认的clone方法为浅拷贝
 * 例：
 *  @classClone
 *	export class MatrixSystem implements System {
 *		roundCount: number;
 *		@write([E1, C1], [E1, C2]) // E1为实体类型， C1， C2为组件类型
 *		run(read: any, write: [Multi<E1, C1>, Multi<E1, C2>]) {
 *			console.log("MatrixSystem run ok");
 *		}
 *	}

 *  @classClone
 *	export class BuffGroupC extends Component {
 *		actionBuff: number;
 *		@deepClone
 *		heap: Heap<Buff>;
 *		@deepClone
 *		roundHeap: Heap<Buff>;
 *		@mapClone
 *		map: Map<string, string>;
 *		@defaultClone([])
 *		instant: Buff[] = [];
 *	}
 */
export const classClone = (target: any): any => {
	let func = (obj: Object) => {
		return Object.create(obj);
	}
	if(target.prototype.__deep_clone__) {
		let f = func;
		func = (obj: Object) => {
			let clone = f(obj);
			for(let r of target.prototype.__deep_clone__) {
				let v = clone[r];
				if(v !== undefined) {
					clone[r] = cloneObject(v);
				}
			}
			return clone
		}
	}
	if(target.prototype.__any_clone__) {
		let f = func;
		func = (obj: Object) => {
			let clone = f(obj);
			for(let r of target.prototype.__any_clone__) {
				clone[r] = cloneAny(clone[r]);
			}
			return clone
		}
	}
	if(target.prototype.__array_clone__) {
		let f = func;
		func = (obj: Object) => {
			let clone = f(obj);
			for(let r of target.prototype.__array_clone__) {
				let v = clone[r];
				if(v !== undefined) {
					clone[r] = [...v];
				}
			}
			return clone
		}
	}
	if(target.prototype.__map_clone__) {
		let f = func;
		func = (obj: Object) => {
			let clone = f(obj);
			for(let r of target.prototype.__map_clone__) {
				let v = clone[r];
				if(v !== undefined) {
					clone[r] = new Map(v.entries());
				}
			}
			return clone
		}
	}
	if(target.prototype.__set_clone__) {
		let f = func;
		func = (obj: Object) => {
			let clone = f(obj);
			for(let r of target.prototype.__set_clone__) {
				let v = clone[r];
				if(v !== undefined) {
					clone[r] = new Set(v.values());
				}
			}
			return clone
		}
	}
	if(target.prototype.__default_clone__) {
		let f = func;
		func = (obj: Object) => {
			let clone = f(obj);
			for(let [k, v] of target.prototype.__default_clone__) {
				clone[k] = v;
			}
			return clone
		}
	}
	target.prototype.clone = function() {
		return func(this);
	}
}
/**
 * 数组的浅拷贝，属性装饰器
 */
export const arrayClone = (target: any, propertyKey: string, _descriptor?: PropertyDescriptor): any => {
	if(target.__array_clone__) {
		target.__array_clone__.push(propertyKey);
	}else{
		target.__array_clone__ = [propertyKey];
	}
}
/**
 * Map的浅拷贝，属性装饰器
 */
export const mapClone = (target: any, propertyKey: string, _descriptor?: PropertyDescriptor): any => {
	if(target.__map_clone__) {
		target.__map_clone__.push(propertyKey);
	}else{
		target.__map_clone__ = [propertyKey];
	}
}
/**
 * Set的浅拷贝，属性装饰器
 */
export const setClone = (target: any, propertyKey: string, _descriptor?: PropertyDescriptor): any => {
	if(target.__set_clone__) {
		target.__set_clone__.push(propertyKey);
	}else{
		target.__set_clone__ = [propertyKey];
	}
}
/**
 * 对象深拷贝，属性装饰器，属性必须为对象，递归将对象深拷贝， 如果对象支持clone方法则会调用clone， 如果为数组、Map、Set, 也会将其中的对象深拷贝
 */
export const deepClone = (target: any, propertyKey: string, _descriptor?: PropertyDescriptor): any => {
	if(target.__deep_clone__) {
		target.__deep_clone__.push(propertyKey);
	}else{
		target.__deep_clone__ = [propertyKey];
	}
}
/**
 * any深拷贝，属性装饰器，属性必须为any，递归将对象深拷贝， 如果对象支持clone方法则会调用clone， 如果为数组、Map、Set, 也会将其中的对象深拷贝
 */
export const anyClone = (target: any, propertyKey: string, _descriptor?: PropertyDescriptor): any => {
	if(target.__any_clone__) {
		target.__any_clone__.push(propertyKey);
	}else{
		target.__any_clone__ = [propertyKey];
	}
}
/**
 * 默认值拷贝，属性装饰器，拷贝时将属性设置为默认值
 */
export const defaultClone = (value: any): any => {
	return function(target: any, propertyKey: string, _descriptor: PropertyDescriptor){
		if(target.__default_clone__) {
			target.__default_clone__.push([propertyKey, value]);
		}else{
			target.__default_clone__ = [[propertyKey, value]];
		}
	}
}
/**
 * 克隆any方法， 如果是对象则调用cloneObject，否则返回原值
 */
export const cloneAny = (value: any): any => {
	return typeof value === 'object' ? cloneObject(value) : value
}
/**
 * 克隆对象方法， 包括null
 */
export const cloneObject = (value: Object): Object => {
	if(value === null) {
		return null
	}
	if(Array.isArray(value)) {
		let clone = [...value];
		if(typeof clone[0] === 'object') {
			for(let i = 0; i < clone.length; i++) {
				clone[i] = cloneObject(clone[i])
			}
		}
		return clone;
	}
	if(value instanceof Map) {
		let first = value.values().next().value;
		if(typeof first === 'object') {
			let clone = new Map;
			for(let [k, v] of value.entries()) {
				clone.set(k, cloneObject(v));
			}
			return clone;
		}else{
			return new Map(value.entries());
		}
	}
	if(value instanceof Set) {
		let first = value.values().next().value;
		if(typeof first === 'object') {
			let clone = new Set;
			for(let v of value.values()) {
				clone.add(cloneObject(v));
			}
			return clone;
		}else{
			return new Set(value.values());
		}
	}
	if (value instanceof ArrayBuffer) {
		return value.slice(0);
	}
	if (ArrayBuffer.isView(value) && (<any>value).BYTES_PER_ELEMENT > 0) {
		return (<any>value).slice(0);
	}
	if(typeof (value as Clone).clone === 'function') {
		return (value as Clone).clone()
	}
	let clone = Object.create(value);
	for (const [k, v] of Object.entries(value)) {
		if(typeof v === 'object') {
			clone[k] = cloneObject(v);
		}else{
			clone[k] = v;
		}
	}
	return clone;
}