/**
 * 小堆
 * 支持删除和更新
 * 
 * @packageDocumentation
 */

// ============================== 导出

/**
 * 小堆
 */
export class Heap<T> {

    /**
     * @description 堆的实际数组，用来表示完全二叉树
     */
    private array: T[] = [];

    /**
     * @description 比较函数 compare Function
     * @param a 元素a
     * @param b 元素b
     */
    private cmp: (a: T, b: T) => number;

    /**
     * 
     * @param cmp 数据的比较函数
     */
    constructor(cmp: (a: T, b: T) => number) {
        this.cmp = cmp;
    }

    /**
     * @description 判空
     */
    public empty(): boolean {
        return this.array.length === 0;
    }

    /**
     * @description 返回内部数组
     */
    public getImpl(): T[] {
        return this.array;
    }

    /**
     * @description 插入元素
     * @param value 目标元素
     */
    public insert(value: T) {
        this.array.push(value);
        this.up(this.array.length - 1);
    }

    /**
     * @description 查找指定元素的位置
     * @param predicate 查找逻辑 find calls predicate once for each element of the array, in ascending order, until it finds one where predicate returns true. If such an element is found, findIndex immediately returns that element index. Otherwise, findIndex returns -1.
     * @param thisArg If provided, it will be used as the this value for each invocation of predicate. If it is not provided, undefined is used instead.
     */
    public findIndex(predicate: (value: T, index: number, obj: T[]) => unknown, thisArg?: any) {
        return this.array.findIndex(predicate, thisArg);
    }
    /**
     * @description 删除元素
     * @param value 目标元素
     */
    public remove(value: T) {
        let index = this.array.indexOf(value);
        index >= 0 && this.removeByIndex(index);
    }
    /**
     * @description 删除指定位置的元素
     * @param index 元素位置
     */
    public removeByIndex(index: number): any {
        // 把最后的叶子赋值给index位置
        const len = this.array.length - 1;
        if (index > len) return;
        const r = this.array[index];
        if(len > 0) {
            this.array[index] = this.array[len];
            this.down(index, len);
        }
        this.array.length = len;
        return r
    }
    /**
     * @description 用过滤函数清除不要的元素， 过滤函数返回true表示保留
     * @param f 过滤函数
     */
    public filter(f: (e: T) => boolean) {
        for(let i = this.array.length - 1; i >= 0; i--) {
            if(!f(this.array[i])) {
                this.removeByIndex(i);
            };
        }
    }

    /**
     * @description 获得但不弹出堆顶元素
     */
    public get(): T {
        return this.array[0];
    }

    /**
     * @description 删除堆顶元素并返回
     */
    public pop(): T {
        return this.removeByIndex(0)
    }

    /**
     * @description 清空
     */
    public clear() {
        this.array.length = 0;
    }

    /**
     * @description 下沉
     * @param curr 
     * @param len
     */
    private down(curr: number, len: number) {
        const arr = this.array;
        const element = arr[curr];
        let left = curr * 2 + 1;
        let right = left + 1;
        while (left < len) {
            // 选择左右孩子的最小值作为比较
            let child = left;
            let e = arr[left];
            if (right < len && this.cmp(arr[right], e) < 0) {
                child = right;
                e = arr[right];
            }

            // 待选择的值比孩子大，则将孩子移到当前的槽
            if (this.cmp(element, e) <= 0) {
                break;
            } else {
                arr[curr] = e;

                // 往下迭代
                curr = child;
                left = curr * 2 + 1;
                right = left + 1;
            }
        }
        arr[curr] = element;
    }

    /**
     * @description 上朔
     * @param curr
     */
    private up(curr: number) {
        const arr = this.array;
        const element = arr[curr];
        let parent = Math.floor((curr - 1) / 2);
        while (parent >= 0) {
            let e = arr[parent];
            if(this.cmp(element, e) >= 0) {
                break
            }
            arr[curr] = e;
            // 往上迭代
            curr = parent;
            parent = Math.floor((curr - 1) / 2);
        }
        arr[curr] = element;
    }
}